from typing import List

from petisco.events.event import Event


class AggregateRoot:
    def __init__(self):
        self.domain_events: List[Event] = []

    def record(self, event: Event):
        self.domain_events.append(event)

    def clear_domain_events(self):
        self.domain_events = []

    def pull_domain_events(self) -> List[Event]:
        return self.domain_events

    def pull_first_domain_event(self) -> Event:
        return self.domain_events[0] if len(self.domain_events) > 0 else None

    def pull_last_domain_event(self) -> Event:
        return self.domain_events[-1] if len(self.domain_events) > 0 else None



from functools import wraps
import traceback

from meiga import Result, isFailure
from meiga.decorators import meiga

from petisco.events.event import Event
from petisco.logger.interface_logger import ERROR, INFO
from petisco.logger.log_message import LogMessage
from petisco.logger.not_implemented_logger import NotImplementedLogger


class _CommandHandler:
    def __init__(self, logger=NotImplementedLogger()):
        self.logger = logger

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):
            @meiga
            def run_command(*args, **kwargs) -> Result:
                return func(*args, **kwargs)

            event = kwargs.get("event")
            if not event or not issubclass(event.__class__, Event):
                raise TypeError(
                    "@command_handler only decorates function with at least one event as input"
                )

            log_message = LogMessage(
                layer="command",
                operation=f"{func.__name__}",
                correlation_id=event.event_correlation_id,
            )

            try:
                log_message.message = f"Event received: {event}"
                self.logger.log(INFO, log_message.to_json())
                result = run_command(*args, **kwargs)
                if result.is_success:
                    log_message.message = f"{result}"
                    self.logger.log(INFO, log_message.to_json())
                else:
                    log_message.message = f"{result}"
                    self.logger.log(ERROR, log_message.to_json())
                return result

            except Exception as e:
                log_message.message = f"Error {func.__name__}: {repr(e.__class__)} {e} | {traceback.format_exc()}"
                self.logger.log(ERROR, log_message.to_json())
                return isFailure

        return wrapper


command_handler = _CommandHandler



from abc import ABCMeta, abstractmethod

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0


class ILogger:

    __metaclass__ = ABCMeta

    @abstractmethod
    def log(self, logging_level: int, message: str):
        return NotImplementedError



import base64
import os
from typing import Any

from meiga import Result, Failure, Error, Success

from petisco.domain.value_objects.value_object import ValueObject
from petisco.domain.errors.input_exceed_lenght_limit_error import (
    InputExceedLengthLimitError,
)

LENGTH = 16


class UserId(str, ValueObject):
    def __new__(cls, user_id, length=LENGTH):
        user_id = None if user_id == "None" else user_id
        cls.length = length
        return str.__new__(cls, user_id)

    def to_result(self) -> Result[Any, Error]:
        user_id = None if self == "None" else self

        if user_id is not None and len(user_id) > self.length:
            return Failure(InputExceedLengthLimitError(message=user_id))
        else:
            return Success(user_id)

    @staticmethod
    def generate():
        r = os.urandom(LENGTH)
        return UserId(base64.b64encode(r, altchars=b"-_").decode("utf-8")[:LENGTH])



# coding=utf-8
# Copyright (C) 2019+ Alice, Vigo, Spain

"""Public API of ALiCE Petisco Framework"""

# Classes
from petisco.application.application_config import ApplicationConfig
from petisco.application.repository import Repository
from petisco.application.service import Service
from petisco.application.singleton import Singleton
from petisco.commands.command_handler import command_handler
from petisco.controller.correlation_id import CorrelationId
from petisco.controller.tokens.jwt_config import JwtConfig
from petisco.domain.aggregate_roots.aggregate_root import AggregateRoot
from petisco.domain.value_objects.client_id import ClientId
from petisco.domain.value_objects.name import Name
from petisco.domain.value_objects.user_id import UserId
from petisco.domain.value_objects.value_object import ValueObject
from petisco.events.event_config import EventConfig
from petisco.logger.logging_based_logger import LoggingBasedLogger
from petisco.logger.interface_logger import ILogger
from petisco.persistence.interface_persistence_connector import IPersistenceConnector
from petisco.use_case import UseCase
from petisco.use_case import use_case_handler
from petisco.controller.controller_handler import controller_handler
from petisco.controller.errors.http_error import HttpError
from petisco.events.event import Event
from petisco.events.event_id import EventId
from petisco.events.interface_event_manager import IEventManager
from petisco.events.not_implemented_event_manager import NotImplementedEventManager

classes = [
    "ApplicationConfig",
    "Service",
    "Repository",
    "UseCase",
    "use_case_handler",
    "controller_handler",
    "command_handler",
    "CorrelationId",
    "HttpError",
    "LoggingBasedLogger",
    "ILogger",
    "Event",
    "EventId",
    "IEventManager",
    "NotImplementedEventManager",
    "Singleton",
    "JwtConfig",
    "IPersistenceConnector",
    "ValueObject",
    "Name",
    "ClientId",
    "UserId",
    "EventConfig",
    "AggregateRoot",
]
# Constants
from petisco.logger.interface_logger import (
    CRITICAL,
    FATAL,
    ERROR,
    WARNING,
    WARN,
    INFO,
    DEBUG,
    NOTSET,
)

constants = ["CRITICAL", "FATAL", "ERROR", "WARNING", "WARN", "INFO", "DEBUG", "NOTSET"]

# Flask
try:
    from petisco.frameworks.flask.application.flask_application import FlaskApplication

    flask = ["FlaskApplication"]
except (RuntimeError, ImportError):
    flask = []

# SqlAlchemy
try:
    from petisco.persistence.sqlalchemy.sqlalchemy_persistence import (
        SqlAlchemyPersistence,
    )
    from petisco.persistence.sqlalchemy.sqlalchemy_persistence_config import (
        SqlAlchemyPersistenceConfig,
    )
    from petisco.persistence.sqlalchemy.sqlalchemy_persistence_connector import (
        SqlAlchemyPersistenceConnector,
    )
    from petisco.persistence.sqlalchemy.sqlalchemy_session_scope import session_scope
    from petisco.persistence.sqlalchemy.sqlalchemy_operational_database_error import (
        SqlAlchemyOperationalDatabaseError,
    )

    sqlalchemy = [
        "SqlAlchemyPersistence",
        "SqlAlchemyPersistenceConfig",
        "SqlAlchemyPersistenceConnector",
        "session_scope",
        "SqlAlchemyOperationalDatabaseError",
    ]
except (RuntimeError, ImportError):
    sqlalchemy = []

# Redis
try:
    from petisco.events.redis.redis_event_manager import RedisEventManager
    from petisco.events.redis.event_from_redis_message import (
        event_from_redis_message,
        EventFromRedisMessageConversionError,
    )

    redis = [
        "RedisEventManager",
        "event_from_redis_message",
        "EventFromRedisMessageConversionError",
    ]
except (RuntimeError, ImportError):
    redis = []

# RabbitMq
try:
    from petisco.events.rabbitmq.rabbitmq_event_manager import RabbitMQEventManager

    rabbitmq = ["RabbitMQEventManager"]
except (RuntimeError, ImportError):
    rabbitmq = []

__all__ = classes + constants + flask + sqlalchemy + redis + rabbitmq



from abc import ABCMeta, abstractmethod
from typing import Any

from meiga import Result, Error, NotImplementedMethodError


class ValueObject:

    __metaclass__ = ABCMeta

    @abstractmethod
    def to_result(self) -> Result[Any, Error]:
        return NotImplementedMethodError

    def guard(self):
        return self.to_result().unwrap_or_return()



from contextlib import contextmanager
from sqlalchemy.exc import OperationalError

from petisco.persistence.sqlalchemy.sqlalchemy_operational_database_error import (
    SqlAlchemyOperationalDatabaseError,
)
from petisco.persistence.sqlalchemy.sqlalchemy_persistence import SqlAlchemyPersistence


@contextmanager
def session_scope():
    """Provide a transactional scope around a series of operations."""
    transactional_scope = SqlAlchemyPersistence.get_instance().session()
    try:
        yield transactional_scope
        transactional_scope.commit()
    except OperationalError:
        transactional_scope.rollback()
        raise SqlAlchemyOperationalDatabaseError
    except Exception as e:
        transactional_scope.rollback()
        raise e
    finally:
        transactional_scope.close()



import jwt


class JwtTokenBuilder:
    @staticmethod
    def build(key, token_type, client_id, issuer="issuer-petisco", user_id=None):
        payload = {
            "iss": issuer,
            "typ": token_type,
            "exp": 1562338273,
            "iat": 1562334673,
            "cli": client_id,
            "sub": user_id,
        }
        token = jwt.encode(payload, key, algorithm="RS256").decode("utf-8")
        return token



import inspect
from functools import wraps
import traceback
from typing import Callable, Tuple, Dict, List, Any

from meiga import Result
from meiga.decorators import meiga

from petisco.application.application_config import ApplicationConfig
from petisco.controller.errors.internal_http_error import InternalHttpError
from petisco.controller.errors.known_result_failure_handler import (
    KnownResultFailureHandler,
)
from petisco.controller.tokens.jwt_decorator import jwt
from petisco.events.request_responded import RequestResponded
from petisco.events.event_config import EventConfig
from petisco.frameworks.flask.headers_provider import flask_headers_provider
from petisco.logger.interface_logger import ERROR, INFO
from petisco.controller.errors.http_error import HttpError
from petisco.controller.tokens.jwt_config import JwtConfig
from petisco.frameworks.flask.correlation_id_provider import (
    flask_correlation_id_provider,
)
from petisco.logger.log_message import LogMessage
from petisco.logger.not_implemented_logger import NotImplementedLogger
from petisco.tools.timer import timer

DEFAULT_SUCCESS_MESSAGE = {"message": "OK"}, 200
DEFAULT_ERROR_MESSAGE = HttpError().handle()


class _ControllerHandler:
    def set_application_config(self, application_config: ApplicationConfig):
        if application_config:
            self.app_name = application_config.app_name
            self.app_version = application_config.app_version
            self.logger = application_config.logger
            self.event_config.event_manager = application_config.event_manager

    def __init__(
        self,
        app_name: str = "app-undefined",
        app_version: str = None,
        logger=NotImplementedLogger(),
        event_config: EventConfig = EventConfig(),
        jwt_config: JwtConfig = None,
        success_handler: Callable[[Result], Tuple[Dict, int]] = None,
        error_handler: Callable[[Result], HttpError] = None,
        correlation_id_provider: Callable = flask_correlation_id_provider,
        headers_provider: Callable = flask_headers_provider,
        logging_types_blacklist: List[Any] = [bytes],
        application_config: ApplicationConfig = None,
    ):
        """
        Parameters
        ----------
        app_name
            Application Name
        app_version
            Application Version
        logger
            A ILogger implementation. Default NotImplementedLogger
        event_config
            EventConfig object. Here, you can define event management.
        jwt_config
            JwtConfig object. Here, you can define how to deal with JWT Tokens
        success_handler
            Handler to deal with Success Results
        error_handler
            Handler to deal with Failure Results
        correlation_id_provider
            Injectable function to provide correlation_id. By default is used flask_correlation_id_provider
        headers_provider
            Injectable function to provide headers. By default is used headers_provider
        logging_types_blacklist
            Logging Blacklist. Object of defined Type will not be logged. By default ( [bytes] ) bytes object won't be logged.
        application_config
            Use ApplicationConfig to set params as: app_name, app_version, logger, or event_manager (EventConfig)
        """
        self.app_name = app_name
        self.app_version = app_version
        self.logger = logger
        self.event_config = event_config
        self.jwt_config = jwt_config
        self.success_handler = success_handler
        self.error_handler = error_handler
        self.correlation_id_provider = correlation_id_provider
        self.headers_provider = headers_provider
        self.logging_types_blacklist = logging_types_blacklist
        self.set_application_config(application_config)

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):
            @timer
            @meiga
            @jwt(jwt_config=self.jwt_config)
            def run_controller(*args, **kwargs) -> Result:
                return func(*args, **kwargs)

            kwargs = get_headers_id(kwargs)
            correlation_id, kwargs = update_correlation_id(kwargs)

            log_message = LogMessage(
                layer="controller",
                operation=f"{func.__name__}",
                correlation_id=correlation_id,
            )

            http_response = DEFAULT_ERROR_MESSAGE
            is_success = False
            elapsed_time = None
            try:
                log_message.message = "Start"
                self.logger.log(INFO, log_message.to_json())
                result, elapsed_time = run_controller(*args, **kwargs)

                if result.is_success:
                    if isinstance(result.value, tuple(self.logging_types_blacklist)):
                        log_message.message = (
                            f"Success result of type: {type(result.value).__name__}"
                        )
                    else:
                        log_message.message = f"{result}"
                    self.logger.log(INFO, log_message.to_json())
                    http_response = (
                        self.success_handler(result)
                        if self.success_handler
                        else DEFAULT_SUCCESS_MESSAGE
                    )
                    is_success = True
                else:
                    log_message.message = f"{result}"
                    self.logger.log(ERROR, log_message.to_json())
                    known_result_failure_handler = KnownResultFailureHandler(result)

                    if not known_result_failure_handler.is_a_result_known_error:
                        if self.error_handler:
                            http_error = self.error_handler(result)

                            if not issubclass(http_error.__class__, HttpError):
                                raise TypeError(
                                    "Returned object from error_handler must be subclasses of HttpError"
                                )

                            http_response = http_error.handle()
                    else:
                        http_response = known_result_failure_handler.http_error.handle()

            except Exception as e:
                log_message.message = f"Error {func.__name__}: {repr(e.__class__)} {e} | {traceback.format_exc()}"
                self.logger.log(ERROR, log_message.to_json())
                http_response = InternalHttpError().handle()

            request_responded = RequestResponded(
                app_name=self.app_name,
                app_version=self.app_version,
                controller=f"{func.__name__}",
                is_success=is_success,
                http_response=http_response,
                correlation_id=correlation_id,
                elapsed_time=elapsed_time,
                additional_info=self.event_config.get_additional_info(kwargs),
            )
            self.event_config.event_manager.publish(
                topic=self.event_config.event_topic, event=request_responded
            )

            return http_response

        def update_correlation_id(kwargs):
            signature = inspect.signature(func)
            if (
                "correlation_id" in signature.parameters
                and "correlation_id" not in kwargs
            ):
                correlation_id = self.correlation_id_provider(func.__name__)
                kwargs = dict(kwargs, correlation_id=correlation_id)
            else:
                correlation_id = None
            return correlation_id, kwargs

        def get_headers_id(kwargs):
            headers = self.headers_provider()
            kwargs = dict(kwargs, headers=headers)

            return kwargs

        return wrapper


controller_handler = _ControllerHandler



import re
from typing import Any

from meiga import Result, Error, Failure, Success

from petisco.domain.value_objects.value_object import ValueObject
from petisco.domain.errors.given_name_is_not_valid_error import GivenNameIsNotValidError
from petisco.domain.errors.input_exceed_lenght_limit_error import (
    InputExceedLengthLimitError,
)


class Name(str, ValueObject):
    def __new__(cls, name, length_limit: int = 50):
        name = None if name == "None" else name
        cls.length_limit = length_limit
        return str.__new__(cls, name)

    def to_result(self) -> Result[Any, Error]:
        name = None if self == "None" else self

        if name is not None:
            if len(self) > self.length_limit:
                return Failure(InputExceedLengthLimitError(message=name))
            else:
                if not re.search(r"^[\w]*(([',. -][\s]?[\w]?)?[\w]*)*$", name):
                    return Failure(GivenNameIsNotValidError(message=name))
        return Success(name)



import re
from typing import Any

from meiga import Result, Error, Failure, Success

from petisco.domain.value_objects.value_object import ValueObject
from petisco.domain.errors.given_input_is_not_valid_error import (
    GivenInputIsNotValidError,
)
from petisco.domain.errors.input_exceed_lenght_limit_error import (
    InputExceedLengthLimitError,
)


class ClientId(str, ValueObject):
    def __new__(cls, client_id, length: int = 50):
        client_id = None if client_id == "None" else client_id
        cls.length = length
        return str.__new__(cls, client_id)

    def to_result(self) -> Result[Any, Error]:
        client_id = None if self == "None" else self

        if client_id is not None:
            if len(client_id) > self.length:
                return Failure(InputExceedLengthLimitError(message=client_id))
            else:
                if not re.search(
                    r"^[a-zA-Z]*(([',. -][a-zA-Z ])?[a-zA-Z]*)*$", client_id
                ):
                    return Failure(GivenInputIsNotValidError(message=client_id))
        return Success(client_id)
